#!/usr/bin/env python3

"""Sadly ipaddress is not very useful by itself"""

import sys
import ipaddress

def int32(i, base=0):
    """IPv4 addresses are 32bit only"""
    return int(i, base) & 0xffffffff

def arpa2ip6(arpa):
    """convert ip6.arpa to ip6"""
    w = arpa.lower().strip('.ip6.arpa').split('.')
    ip = []
    w.reverse()
    x = 0
    n = len(w)
    while x < n:
        ip.append(''.join(w[x:x+4]))
        x += 4
    if len(ip) < 6:
        # we need :: to make it valid
        ip.append('')
        ip.append('')
    ip6 = ipaddress.IPv6Address(':'.join(ip))
    return str(ip6)

def ip2arpa6(ip6):
    """ip6.arpa for ip6

    This is just using ipaddress.IPv6Address
    """
    ip6 = ipaddress.IPv6Address(ip6)
    return ip6.reverse_pointer

def ip2arpa(ip):
    if isinstance(ip, ipaddress.IPv6Address) or ':' in ip:
        return ip2arpa6(ip)
    a = ip.split('.')
    a.reverse()
    return '.'.join(a)+'.in-addr.arpa'

def arpa2ip(arpa, zeros=True):
    if 'ip6.arpa' in arpa:
        return arpa2ip6(arpa)
    a = arpa.lower().strip('.in-addr.arpa').split('.')
    a.reverse()
    if zeros:
        a += ['0','0','0']
        return '.'.join(a[:4])
    return '.'.join(a)
    
def ip2hex(ip):
    a = ipaddress.ip_address(ip)
    return hex(int(a))

def hex2ip(h):
    a = ipaddress.ip_address(h)
    return str(a)

def get_class_prefix(ip, conf={}):
    """return the default class prefix for ip
    with or without trailing 0's
    """
    if isinstance(ip, ipaddress.IPv6Address) or '.' not in ip:
        return ip
    a = ip.split('.')
    c = int(a[0])

    zeros = conf.get('class_prefix_zeros', False)
    if c < 128:
        n = 1
    elif c < 192:
        n = 2
    else:
        n = 3
    k = '.'.join(a[:n])
    if zeros:
        return k + '.' + '.'.join(['0', '0', '0', '0'][n:])
    return k

def get_mask(ip, conf={}):
    """return the netmask for ip

    If a specific mask has been set we give that
    otherwise the default class based mask
    """

    p = get_class_prefix(ip, conf)
    cidir_nets = conf.get('cidr_nets', {})
    if p in cidir_nets:
        # there can be more than on cidr net with
        # the same class prefix
        for c in cidir_nets[p]:
            if c.is_member(ip):
                return hex(c.mask)

    masks = conf.get('masks', {})
    if p in masks:
        return masks[p]

    if isinstance(ip, ipaddress.IPv6Address) or '.' not in ip:
        return None
    c = int(ip.split('.')[0])

    if c < 128:
        dmask = masks.get('A', '0xff000000')
    elif c < 192:
        dmask = masks.get('B', '0xffff0000')
    else:
        dmask = masks.get('C', '0xffffff00')
    return dmask

def get_cidr_net(ip, conf={}):
    """see if we have a cidr_net that ip is a member of"""
    cidr_nets = conf.get('cidr_nets', {})
    if ip.endswith('.arpa'):
        if ip in cidr_nets:
            if len(cidr_nets[ip]) == 1:
                return cidr_nets[ip][0]
        # this usually works fine
        ip = arpa2ip(ip)
    prefix = get_class_prefix(ip, conf)
    if prefix in cidr_nets:
        for c in cidr_nets[prefix]:
            if c.is_member(ip):
                return c
    mask = get_mask(ip, conf)
    c = CIDR(ip, mask, conf)
    return c

class CIDR(object):
    def __init__(self,ip1,ip2=None,conf={}):
        """input can be any of:
        ip1=net[/-]bits
        net,bits
        net,mask
        ip1,ip2
        """
        self.conf = conf
        self.sep = conf.get('cidr_sep', '/')
        self.bits = 0
        self.mask = None
        prefix = ip1

        if isinstance(ip1, ipaddress.IPv6Address):
            ip1 = str(ip1)
            self.is_ip6 = True
        else:
            self.is_ip6 = ':' in ip1
        if self.is_ip6:
            self.max_bits = 128
            self.int_mask = 0xffffffffffffffffffffffffffffffff
        else:
            self.max_bits = 32
            self.int_mask = 0xffffffff

        if not ip2:
            if '/' in ip1:
                self.sep = '/'
            elif '-' in ip1:
                self.sep = '-'
            if self.sep in ip1:         # net[/-]bits
                prefix,bits = ip1.split(self.sep)
                self.bits = int(bits)
            elif self.is_ip6:
                self.bits = 64          # max allowed
        elif self.is_ip6:
            if not (isinstance(ip2, ipaddress.IPv6Address) or ':' in ip2):
                if ip2.startswith('ff') or ip2.startswith('0xff'):
                    self.mask = hex(int(ip2, 16))
                else:
                    self.bits = int(ip2)
        elif '.' not in ip2:            # bits
            if ip2.startswith('ff') or ip2.startswith('0xff'):
                self.mask = hex(int(ip2, 16))
            else:
                self.bits = int(ip2)
        elif ip2.startswith('255.'):
            self.mask = ip2hex(ip2)
        if self.mask:
            self.mask = int(self.mask, 16)
            self.set_bits()
        elif self.bits > 0:
            self.set_mask()

        if self.is_ip6:
            if not self.mask:
                # we must have 2 addresses
                self.a1 = int(ipaddress.IPv6Address(prefix))
                self.a2 = int(ipaddress.IPv6Address(ip2))
                delta = self.a2 - self.a1
                # there is always 64bits of host address
                self.mask = ~(delta | ((1<<64) -1)) & self.int_mask
                self.set_bits()
            else:
                self.a1 = int(ipaddress.IPv6Address(prefix)) & self.mask
                self.a2 = self.a1 + ~self.mask & self.int_mask
            self.host_nibbles = int((self.max_bits - self.bits) / 4)
            self.prefix = self.ip1 = ipaddress.IPv6Address(self.a1)
            self.ip2 = ipaddress.IPv6Address(self.a2)
            self.class_prefix = self.prefix.reverse_pointer
        else:
            if not self.mask:
                # we must have 2 addresses
                self.a1 = int(ip2hex(prefix), 16)
                self.a2 = int(ip2hex(ip2), 16)
                delta = self.a2 - self.a1
                self.mask = ~delta & self.int_mask
                self.set_bits()
            else:
                self.a1 = int(ip2hex(prefix), 16) & self.mask
                self.a2 = self.a1 + ~self.mask & self.int_mask
            self.host_octets = int((self.max_bits - self.bits) / 8)
            self.prefix = self.ip1 = hex2ip(self.a1)
            self.ip2 = hex2ip(self.a2)
            self.class_prefix = get_class_prefix(self.prefix, self.conf)

        if not 'cidr_nets' in self.conf:
            self.conf['cidr_nets'] = {}
        # there can be more than one net with the same class_prefix
        if not self.class_prefix in self.conf['cidr_nets']:
            self.conf['cidr_nets'][self.class_prefix] = []
        self.conf['cidr_nets'][self.class_prefix].append(self)
        if not self.is_ip6:
            in_addr0 = self.in_addr_zone(0)
            if not in_addr0 in self.conf['cidr_nets']:
                self.conf['cidr_nets'][in_addr0] = []
            self.conf['cidr_nets'][in_addr0].append(self)

    def set_mask(self, mask=None):
        if not mask is None:
            self.mask = mask
            return
        self.mask = ~((1 << (self.max_bits - self.bits)) - 1) & self.int_mask
        
    def set_bits(self, bits=None):
        if not bits is None:
            self.bits = bits
            return
        x = ~self.mask & self.int_mask
        y = 1
        for j in range(self.max_bits):
            y |= (1 << j)
            if x == y:
                self.bits = self.max_bits - 1 - j
                break

    def in_addr_zone(self, bits=None):
        if bits != 0:
            bits = self.bits
        if self.is_ip6:
            a = self.class_prefix.split('.')
            return '.'.join(a[self.host_nibbles:])
        a = ip2arpa(self.prefix).split('.')
        if bits % 8 == 0:
            return '.'.join(a[self.host_octets:])
        # handle this per rfc2317
        a[0] = '{}{}{}'.format(a[0], self.sep, self.bits)
        return '.'.join(a)

    def is_member(self, ip):
        """is ip part of this net"""
        a = int(ip2hex(ip), 16)
        if a < self.a1 or a > self.a2:
            return False
        return True

    def __str__(self):
        return '{}{}{} {} {} - {}'.format(self.prefix,
                                          self.sep,
                                          self.bits,
                                          hex(self.mask),
                                          self.ip1, self.ip2)
        
if __name__ == '__main__':
    import getopt
    import conf as cf

    conf = {}
    opts,args = getopt.getopt(sys.argv[1:], 'c:o:')

    for o,a in opts:
        if o == '-c':
            conf = cf.loadConfig(a, conf)
        elif o == '-o':
            conf = cf.stringConfig(a, conf)
            
    for a in args:
        if '/' in a or '-' in a:
            c = CIDR(a, conf=conf)
            print(c)
        elif ',' in a:
            a1,a2 = a.split(',')
            c = CIDR(a1,a2, conf)
            print(c)
        else:
            print('{} {}'.format(a, get_mask(a)))

    cidr_nets = conf.get('cidr_nets', {})
    for p,l in cidr_nets.items():
        if p.endswith('.arpa'):
            continue
        print('class_prefix={}:'.format(p))
        for c in l:
            print(c)
            print(c.in_addr_zone())
            if not c.is_ip6:
                print(c.in_addr_zone(0))
    
